from itertools import permutations
from itertools import product
from typing import Dict
from typing import Tuple

import numpy as np
from typing_extensions import TypeAlias

from avalon.common.utils import float_to_str
from avalon.datagen.world_creation.configs.building import BuildingConfig
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.indoor.builders import DefaultHallwayBuilder
from avalon.datagen.world_creation.indoor.builders import DefaultStoryLinker
from avalon.datagen.world_creation.indoor.builders import HouseLikeRoomBuilder
from avalon.datagen.world_creation.indoor.builders import IrregularFootprintBuilder
from avalon.datagen.world_creation.indoor.builders import RectangleFootprintBuilder
from avalon.datagen.world_creation.indoor.builders import TLShapeFootprintBuilder
from avalon.datagen.world_creation.indoor.builders import WindowBuilder
from avalon.datagen.world_creation.indoor.building import BuildingTask
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.worlds.compositional import get_radius_for_building_task

CANONICAL_SEED = 42
BuildingCatalog = Dict[str, BuildingConfig]
IndoorWorldParams = Tuple[np.random.Generator, float, BuildingTask, float, Point3DNP]
IndoorWorldCatalog = Dict[str, IndoorWorldParams]


def canonical_rand() -> np.random.Generator:
    return np.random.default_rng(CANONICAL_SEED)


def _create_building_catalogs() -> Tuple[BuildingCatalog, ...]:
    size_options = {
        "tiny": dict(width=5, length=5),
        "small": dict(width=8, length=8),
        "regular": dict(width=16, length=16),
        "narrow": dict(width=5, length=17),
        "wide": dict(width=16, length=6),
        "large": dict(width=32, length=32),
    }

    height_options = {
        "1": dict(story_count=1),
        "2_decorative": dict(story_count=2, story_linker=None),
        "2_functional": dict(story_count=2, story_linker=DefaultStoryLinker()),
        "5_decorative": dict(story_count=5, story_linker=None),
        "5_functional": dict(story_count=5, story_linker=DefaultStoryLinker()),
    }
    footprint_options = {
        "rectangle": dict(footprint_builder=RectangleFootprintBuilder()),
        "tl_shape": dict(footprint_builder=TLShapeFootprintBuilder()),
        "irregular": dict(footprint_builder=IrregularFootprintBuilder()),
    }

    other_params = dict(
        room_builder=HouseLikeRoomBuilder(), hallway_builder=DefaultHallwayBuilder(), window_builder=WindowBuilder()
    )

    incompatible_combos = {
        ("tl_shape", "tiny"),
        ("irregular", "tiny"),
    }

    valid_building_catalog = {}
    invalid_building_catalog = {}
    for (size, size_params), (height, height_params), (footprint, footprint_params), rest in product(
        size_options.items(), height_options.items(), footprint_options.items(), [other_params]
    ):
        param_type_permutations = {
            *permutations([size, height, footprint], 2),
            *permutations([size, height, footprint], 3),
        }
        is_combination_valid = len(param_type_permutations.intersection(incompatible_combos)) == 0
        params_id = "__".join([size, height, footprint])
        params = {**size_params, **height_params, **footprint_params, **other_params}  # type: ignore
        if is_combination_valid:
            valid_building_catalog[params_id] = BuildingConfig(**params)  # type: ignore
        else:
            invalid_building_catalog[params_id] = BuildingConfig(**params)  # type: ignore
    return valid_building_catalog, invalid_building_catalog


def _create_indoor_world_catalog() -> IndoorWorldCatalog:
    indoor_tasks = list(BuildingTask)
    indoor_task_difficulties = [round(d, 1) for d in np.arange(0.0, 1.2, 0.2)]
    indoor_world_catalog: IndoorWorldCatalog = {}

    param_rand = canonical_rand()
    for task, difficulty in product(indoor_tasks, indoor_task_difficulties):
        seed = param_rand.integers(0, np.iinfo(np.int64).max)
        task_rand = np.random.default_rng(seed)

        radius = get_radius_for_building_task(task_rand, task, difficulty)
        location = task_rand.uniform(-100, 100, (3,))
        params = task_rand, difficulty, task, radius, location
        world_name = "__".join(["indoor_world", task.value.lower(), float_to_str(difficulty)])
        indoor_world_catalog[world_name] = params
    return indoor_world_catalog


Difficulty: TypeAlias = float
Seed: TypeAlias = int
OutdoorWorldParams = Tuple[AvalonTask, Difficulty, Seed]
OutdoorWorldCatalog = Dict[str, OutdoorWorldParams]


def _create_outdoor_world_catalog() -> OutdoorWorldCatalog:
    tasks = list(AvalonTask)
    difficulties = [round(d, 1) for d in np.arange(0.0, 1.1, 0.1)]
    outdoor_world_catalog: OutdoorWorldCatalog = {}

    for seed, (task, difficulty) in enumerate(product(tasks, difficulties)):
        world_name = "__".join(["outdoor_world", str(seed), task.value.lower(), float_to_str(difficulty)])
        outdoor_world_catalog[world_name] = (task, difficulty, seed)
    return outdoor_world_catalog


VALID_BUILDING_CATALOG, INVALID_BUILDING_CATALOG = _create_building_catalogs()
BUILDING_CATALOG = {**VALID_BUILDING_CATALOG, **INVALID_BUILDING_CATALOG}
INDOOR_WORLD_CATALOG = _create_indoor_world_catalog()
OUTDOOR_WORLD_CATALOG = _create_outdoor_world_catalog()



"""
Code for computing Generalized Advantage Estimation.
"""
import torch
from torch import Tensor


def gae(rewards: Tensor, values: Tensor, dones: Tensor, gamma: float, lam: float, last_values: Tensor) -> Tensor:
    """Compute generalized advantage estimate.
    Expects ordering from rollouts with obs_first=True, ie (obs/value, action, reward/done).
    rewards: a list of rewards at each step.
    values: the value estimate of the state at each step.
    dones: an array of the same shape as rewards, with a 1 if the
        episode ended at that step and a 0 otherwise.
    gamma: the discount factor.
    lam: the GAE lambda parameter. if lam=0, we perform 1-step value boostrapping.
        if lam=1, we use discounted emperical returns (with value backup at the end of the fragment)
    last_values: the values computed from the terminal observation (O_T+1)
    """
    # Invert dones to have 0 if the episode ended and 1 otherwise
    not_dones = (dones * -1) + 1

    N = rewards.shape[0]
    T = rewards.shape[1]
    gae_step = torch.zeros(size=(N,), device=rewards.device, dtype=torch.float32)
    advantages = torch.zeros(size=(N, T), device=rewards.device, dtype=torch.float32)
    for t in reversed(range(T)):
        if t == T - 1:
            next_values = last_values
        else:
            next_values = values[:, t + 1]

        # First compute delta, which is the one-step TD error
        delta = rewards[:, t] + gamma * next_values * not_dones[:, t] - values[:, t]
        # Then compute the current step's GAE by discounting the previous step
        # of GAE, resetting it to zero if the episode ended, and adding this step's delta
        gae_step = delta + gamma * lam * not_dones[:, t] * gae_step
        assert gae_step.shape == (N,)
        # And store it
        advantages[:, t] = gae_step
    return advantages



# %%
import json
import multiprocessing as mp
import os
import tarfile
import traceback
from datetime import datetime
from pathlib import Path
from typing import Dict
from typing import Sequence

import sentry_sdk
import torch

from avalon.agent.godot.godot_gym import AvalonEnv
from avalon.agent.godot.godot_gym import GodotEnvironmentParams
from avalon.common.error_utils import setup_sentry
from avalon.common.log_utils import configure_parent_logging
from avalon.common.log_utils import configure_remote_logger
from avalon.common.log_utils import logger
from avalon.contrib.s3_utils import SimpleS3Client
from avalon.contrib.utils import FILESYSTEM_ROOT
from avalon.datagen.godot_env.actions import VRAction
from avalon.datagen.godot_env.godot_env import GodotEnv

EVAL_TEMP_PATH = "/tmp/avalon_eval"


class RandomActionDistribution:
    def __init__(self, real_log_stds: Sequence[float], discrete_logits: Sequence[float], seed: int) -> None:
        torch.random.manual_seed(seed)
        real_scales = torch.tensor(real_log_stds).exp()
        real_means = torch.zeros_like(real_scales)
        self.real_dist = torch.distributions.Normal(loc=real_means, scale=real_scales)
        self.discrete_dist = torch.distributions.Bernoulli(logits=torch.tensor(discrete_logits))

    def sample(self):
        return {"real": self.real_dist.sample(), "discrete": self.discrete_dist.sample()}


@torch.no_grad()
def do_eval_rollouts(policy: RandomActionDistribution, env: GodotEnv) -> Dict[str, float]:
    # DO ROLLOUTS
    observation = env.reset()
    frame_count = 1
    scores_by_world_index: Dict[str, float] = {}
    while True:
        action = VRAction.from_input(policy.sample())
        observation, goal_progress = env.act(action)
        frame_count += 1
        world_index = goal_progress.log["world_index"]

        if world_index in scores_by_world_index:
            break
        if goal_progress.is_done:
            info = goal_progress.log
            logger.info(
                f"Episode {world_index}, {info['task']} ({info['difficulty']}) score: {info['score']:.3f} in {frame_count} frames"
            )
            scores_by_world_index[world_index] = goal_progress.log["score"]

            env.reset()
            frame_count = 1
    return scores_by_world_index


def run_policy_on_world_ids(
    eval_world_dir: str, process_index: int = 0, num_processes: int = 1, seed: int = 0
) -> Dict[str, float]:
    policy = RandomActionDistribution([0] * 18, [0] * 3, seed=seed)
    params = GodotEnvironmentParams(
        mode="test", fixed_worlds_load_from_path=Path(eval_world_dir), env_index=process_index, env_count=num_processes
    )
    env = AvalonEnv(params)
    result = do_eval_rollouts(policy, env)
    return result


def get_random_result_key(seed: int, data_key: str):
    return "__".join(["random_policy", str(seed), data_key]).replace("/", "_")


def run_evaluation(worlds_path: Path, data_key: str, seed: int) -> None:

    s3_client = SimpleS3Client()

    setup_sentry(percent_of_traces_to_capture=1.0)

    if data_key is None:
        assert worlds_path.is_dir()
    else:
        logger.info(f"Downloading data from {data_key} to {worlds_path}")
        os.makedirs(worlds_path, exist_ok=True)
        os.makedirs(EVAL_TEMP_PATH, exist_ok=True)
        s3_client.download_to_file(data_key, Path(EVAL_TEMP_PATH) / data_key)
        with tarfile.open(Path(EVAL_TEMP_PATH) / data_key, "r:gz") as f:
            f.extractall(worlds_path)

    num_processes = 10
    all_results = {}

    def on_done(result: Dict[str, float]) -> None:
        logger.info(f"Finished generating {result}")
        all_results.update(result)

    def on_error(error: BaseException) -> None:
        logger.error(f"Evaluation failed! {error}")
        traceback.print_exc()
        sentry_sdk.capture_exception(error)

    start_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    ctx = mp.get_context("spawn")
    with ctx.Pool(processes=num_processes, initializer=configure_parent_logging) as worker_pool:
        requests = []
        for process_index in range(num_processes):
            request = worker_pool.apply_async(
                run_policy_on_world_ids,
                kwds={
                    "eval_world_dir": str(worlds_path),
                    "process_index": process_index,
                    "num_processes": num_processes,
                    "seed": seed,
                },
                callback=on_done,
                error_callback=on_error,
            )
            requests.append(request)
        worker_pool.close()
        worker_pool.join()
    end_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    record_data = {
        "baseline": "random",
        "data_key": data_key,
        "eval_world_dir": str(worlds_path),
        "start_time": start_time,
        "end_time": end_time,
        "all_results": all_results,
    }

    result_key = get_random_result_key(seed, data_key)
    logger.info(f"Saving result to '{result_key}'")
    s3_client.save(result_key, json.dumps(record_data))  # type: ignore


if __name__ == "__main__":
    configure_remote_logger()
    seed = 0
    worlds_path = Path(f"{FILESYSTEM_ROOT}/avalon/worlds/viewable_worlds")
    data_key = "avalon_worlds__0824_full.tar.gz"
    run_evaluation(worlds_path=worlds_path, data_key=data_key, seed=seed)



from itertools import product
from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

import attr
import godot_parser
import numpy as np
from godot_parser import ExtResource
from godot_parser import Node as GDNode
from nptyping import Float
from nptyping import NDArray
from nptyping import Shape
from scipy.spatial.transform import Rotation

from avalon.datagen.godot_base_types import FloatRange
from avalon.datagen.godot_base_types import IntRange
from avalon.datagen.world_creation.indoor.constants import DEFAULT_FLOOR_THICKNESS
from avalon.datagen.world_creation.indoor.constants import DEFAULT_STORY_HEIGHT
from avalon.datagen.world_creation.types import GDLinkedSection
from avalon.datagen.world_creation.types import GodotScene
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.utils import make_transform

RangeType = TypeVar("RangeType")
LBT = TypeVar("LBT", bound="LevelBlock")


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Block(Generic[RangeType]):
    x: RangeType
    y: RangeType
    z: RangeType


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class LevelBlock(Block):
    """
    A representation of a box-like shape to be used in an Avalon level context - these can then later be converted to
    CubeMeshes, BoxShapes, and any other rectangle prisms in Godot.

    Blocks are rotated around their origin (centroid).
    """

    rotation: Rotation = Rotation.identity()
    is_visual: bool = True
    is_collider: bool = True
    is_climbable: bool = False

    def __attrs_post_init__(self) -> None:
        assert self.is_visual or self.is_collider, "LevelBlocks must be either colliders, visual or both"

    @classmethod
    def make(
        cls: Type[LBT], centroid: Point3DNP, size: Point3DNP, rotation: Rotation = Rotation.identity(), **kwargs: Any
    ) -> LBT:
        x = FloatRange(centroid[0] - size[0] / 2, centroid[0] + size[0] / 2)
        y = FloatRange(centroid[1] - size[1] / 2, centroid[1] + size[1] / 2)
        z = FloatRange(centroid[2] - size[2] / 2, centroid[2] + size[2] / 2)
        rotation = rotation
        return cls(x, y, z, rotation, **kwargs)

    @property
    def type_name(self) -> str:
        return self.__class__.__name__.lower().removesuffix("block")

    @property
    def centroid(self) -> Tuple[float, float, float]:
        return self.x.midpoint, self.y.midpoint, self.z.midpoint

    @property
    def size(self) -> Tuple[float, float, float]:
        return self.x.size, self.y.size, self.z.size


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class FloorBlock(LevelBlock):
    pass


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class CeilingBlock(LevelBlock):
    pass


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class WindowBlock(LevelBlock):
    is_visual: bool = False


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class WallBlock(LevelBlock):
    is_interior: bool = True

    @property
    def type_name(self) -> str:
        return f"{('interior' if self.is_interior else 'exterior')}_{super().type_name}"


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class LadderBlock(LevelBlock):
    is_climbable: bool = True


def make_blocks_from_tiles(grid: np.ndarray, solid_tile_value: Any = 0) -> List[Block]:
    """
    Divide a grid of `grid` into as few solid `Block`s as possible. Blocks represent rectangular regions of the grid
    that all consist of solid tiles. Tiles are considered to be solid if their value equals `solid_tile_value`.

    The algorithm iterates tile-by-tile to find a solid tile that is not already part of a block. It then starts a new
    block. The block formation goes greedily left-to-right to find a row of solid tiles and then extends the row
    downward as far as possible.
    """

    blocks = []
    covered_tiles: Set[Tuple[int, int]] = set()
    for (z, x), _tile in np.ndenumerate(grid):
        if (z, x) in covered_tiles or grid[z][x] != solid_tile_value:
            continue
        block = _extract_next_block(grid, z, x, covered_tiles, solid_tile_value)
        if block is not None:
            blocks.append(block)
    return blocks


def _extract_next_block(
    grid: np.ndarray, z_min_ge: int, x_min_ge: int, covered_tiles: Set[Tuple[int, int]], solid_tile_value: int = 0
) -> Block:
    length, width = grid.shape

    # Extend along x
    xi = x_min_ge
    for xi in range(x_min_ge, width + 1):
        if xi >= width or grid[z_min_ge][xi] != solid_tile_value or (z_min_ge, xi) in covered_tiles:
            # This tile is outside grid, not solid or already covered, so this is the furthest we go along x
            break
    x_max_lt = xi

    # Extend along z
    zi = z_min_ge
    for zi in range(z_min_ge, length + 1):
        if zi >= length:
            # We've reached the end of the grid
            break

        candidate_tiles = grid[zi : zi + 1, x_min_ge:x_max_lt]
        if (candidate_tiles == solid_tile_value).all():
            candidate_coords = set(product(range(zi, zi + 1), range(x_min_ge, x_max_lt)))
            has_dupe_tiles = len(candidate_coords.intersection(covered_tiles)) > 0
            if has_dupe_tiles:
                break
        else:
            break
    z_max_lt = zi

    # Make sure we don't create overlapping blocks by marking the tiles as covered
    covered_tiles.update(product(range(z_min_ge, z_max_lt), range(x_min_ge, x_max_lt)))

    return Block(
        x=IntRange(x_min_ge, x_max_lt),
        y=IntRange(DEFAULT_FLOOR_THICKNESS, DEFAULT_STORY_HEIGHT + 1),
        z=IntRange(z_min_ge, z_max_lt),
    )


def new_make_block_nodes(
    scene: GodotScene,
    position: NDArray[Shape["3,0"], Float],
    size: NDArray[Shape["3,0"], Float],
    rotation: NDArray[Shape["12,0"], Float] = np.array([1, 0, 0, 0, 1, 0, 0, 0, 1]),
    make_parent: bool = True,
    parent_type: str = "StaticBody",
    parent_name: str = "StaticBody",
    parent_script: Optional[ExtResource] = None,
    parent_extra_props: Optional[Dict[str, Any]] = None,
    make_mesh: bool = True,
    mesh_name: str = "mesh",
    mesh_material: Optional[Union[str, GDLinkedSection]] = None,
    mesh_script: Optional[ExtResource] = None,
    make_collision_shape: bool = True,
    collision_shape_name: str = "collision_shape",
    collision_shape_script: Optional[ExtResource] = None,
    collision_shape_disabled: bool = False,
) -> List[GDNode]:
    nodes = []
    width, height, length = size

    transform = make_transform(position=position, rotation=rotation)

    if make_mesh:
        assert mesh_material is not None
        if isinstance(mesh_material, str):
            mesh_material = scene.add_ext_resource(mesh_material, "Material")
        mesh_props = {
            "mesh": scene.add_sub_resource("CubeMesh", size=godot_parser.Vector3(width, height, length)).reference,
            "material/0": mesh_material.reference,
        }
        if not make_parent:
            mesh_props["transform"] = transform
        if mesh_script:
            mesh_props["script"] = mesh_script
        mesh_instance = GDNode(mesh_name, type="MeshInstance", properties=mesh_props)
        nodes.append(mesh_instance)

    if make_collision_shape:
        collision_shape_props = {}
        if not make_parent:
            collision_shape_props["transform"] = transform

        collision_shape_props["shape"] = scene.add_sub_resource(
            "BoxShape",
            extents=godot_parser.Vector3(width / 2, height / 2, length / 2),
        ).reference

        if collision_shape_script:
            collision_shape_props["script"] = collision_shape_script
        if collision_shape_disabled:
            collision_shape_props["disabled"] = True
        collision_shape = GDNode(collision_shape_name, type="CollisionShape", properties=collision_shape_props)
        if make_collision_shape:
            nodes.append(collision_shape)

    if make_parent:
        parent_props = {}
        if parent_script:
            parent_props["script"] = parent_script
        parent_props.update({} if parent_extra_props is None else parent_extra_props)
        parent_props.update(transform=transform)
        parent_node = GDNode(parent_name, parent_type, properties=parent_props)
        for node in nodes:
            parent_node.add_child(node)
        nodes = [parent_node]

    return nodes


BlocksByStory = Dict[int, List[LevelBlock]]



from pathlib import Path
from typing import Optional
from typing import Tuple
from typing import Type

import attr
import numpy as np
from scipy import stats

from avalon.datagen.errors import WorldTooSmall
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import CLIMBING_REQUIRED_HEIGHT
from avalon.datagen.world_creation.constants import ITEM_FLATTEN_RADIUS
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.entities.animals import Predator
from avalon.datagen.world_creation.entities.constants import LARGEST_ANIMAL_SIZE
from avalon.datagen.world_creation.tasks.avoid import select_predator_types
from avalon.datagen.world_creation.tasks.eat import add_food_tree_for_simple_task
from avalon.datagen.world_creation.tasks.hunt import ForceWeapons
from avalon.datagen.world_creation.tasks.hunt import select_weapons
from avalon.datagen.world_creation.worlds.creation import create_world_from_constraint
from avalon.datagen.world_creation.worlds.difficulty import difficulty_variation
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.export import export_world
from avalon.datagen.world_creation.worlds.obstacles.configure import create_inner_placeholder_solution
from avalon.datagen.world_creation.worlds.obstacles.configure import make_ring
from avalon.datagen.world_creation.worlds.obstacles.height_path import HeightPath
from avalon.datagen.world_creation.worlds.obstacles.height_solution import HeightSolution
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class FightTaskConfig(TaskConfig):
    # how many predators to have at difficulty 0.0 and 1.0 respectively. Note that each one will be a different type.
    predator_count_easy: int = 1
    predator_count_hard: int = 5
    # how far the predators will be moved from their canonical location (ie, directly in the way) at difficulty 0.0
    # and 1.0 respectively.
    predator_randomization_dist_easy: float = 10.0
    predator_randomization_dist_hard: float = 3.0
    # how far away weapons will spawn. At max difficulty, they will spanw randomly between the min and max
    # At the lowest difficulty, they will be within the min radius
    weapon_radius_min: float = 2.0
    weapon_radius_max: float = 5.0
    # how far away our goal should be at difficulty 0.0 and 1.0 respectively
    # there needs to be enough room so that you dont start right next to the predators and get killed instantly
    goal_dist_easy: float = 12.0
    goal_dist_hard: float = 24.0
    # how long to make the corridor which requires you to actually fight enemies (on max difficulty)
    # on easier difficulties it will end up being a bit shorter (and may be shorter if the world is a bit too small
    # for a longer one to fit)
    path_length_max: float = 6.0
    # how wide to make the corridor in which the animals spawn at difficulty 0.0 and 1.0 respectively
    # controls how easy it is to avoid them. Is normally distributed with the mean based on the difficulty
    path_width_easy: float = 10.0
    path_width_hard: float = 2.5
    path_width_std_dev: float = 0.5
    # how tall the walls are. Always min height at difficulty 0.0, uniformly between the two at difficulty 1.0
    height_min: float = CLIMBING_REQUIRED_HEIGHT * 1.25
    height_max: float = CLIMBING_REQUIRED_HEIGHT * 2.0
    # these parameters control which weapons will be spawned
    # the probability that any given weapons will be a rock (vs a stick)
    rock_probability: float = 0.5
    # some weapons are more useful than others. They are spawned until enough "value" has been spawned
    weapon_value_easy: float = 5.0
    weapon_value_hard: float = 2.0
    weapon_value_std_dev: float = 0.5
    # there are large variants of weapons, which do more damage. They become less likely to spawn at higher difficulties
    large_weapon_probability_easy: float = 0.8
    large_weapon_probability_hard: float = 0.2


@attr.define
class ForceFight(ForceWeapons):
    predators: Optional[Tuple[Type[Predator], ...]] = None
    predator_randomization_radius: Optional[float] = None
    gap_distance: Optional[float] = None
    path_width: Optional[float] = None
    desired_goal_distance: Optional[float] = None


def generate_fight_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: FightTaskConfig = FightTaskConfig(),
    _FORCED: Optional[ForceFight] = None,
) -> None:
    world, locations, difficulty = create_fight_obstacle(
        rand, difficulty, export_config, _forced_fight=_FORCED, task_config=task_config
    )
    world, locations = world.end_height_obstacles(
        locations, is_accessible_from_water=False, is_spawn_region_climbable=False
    )
    world = add_food_tree_for_simple_task(world, locations)
    world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal)
    export_world(output_path, rand, world)


def create_fight_obstacle(
    rand: np.random.Generator,
    difficulty: float,
    export_config: ExportConfig,
    constraint: Optional[CompositionalConstraint] = None,
    is_for_carry: bool = False,
    _forced_fight: Optional[ForceFight] = None,
    task_config: FightTaskConfig = FightTaskConfig(),
) -> Tuple[World, WorldLocations, float]:
    if _forced_fight is None:
        _forced_fight = ForceFight()

    predator_types, difficulty = select_predator_types(
        task_config.predator_count_easy,
        task_config.predator_count_hard,
        0.5,
        difficulty,
        rand,
        _FORCED=_forced_fight.predators,
    )

    # need some time and space to prepare...
    desired_goal_distance = scale_with_difficulty(
        difficulty, task_config.goal_dist_easy, task_config.goal_dist_hard, _FORCED=_forced_fight.desired_goal_distance
    )

    world, locations = create_world_from_constraint(
        stats.norm(desired_goal_distance, 0.5), rand, difficulty, export_config, constraint
    )

    predator_randomization_radius = scale_with_difficulty(
        difficulty,
        task_config.predator_randomization_dist_easy,
        task_config.predator_randomization_dist_hard,
        _FORCED=_forced_fight.predator_randomization_radius,
    )

    weapons, difficulty = select_weapons(
        rand,
        difficulty,
        is_rock_necessary=False,
        rock_probability=_forced_fight.rock_probability or task_config.rock_probability,
        weapon_value_easy=task_config.weapon_value_easy,
        weapon_value_hard=task_config.weapon_value_hard,
        weapon_value_std_dev=task_config.weapon_value_std_dev,
        large_weapon_probability_easy=task_config.large_weapon_probability_easy,
        large_weapon_probability_hard=task_config.large_weapon_probability_hard,
        _FORCED_WEAPON_VALUE=_forced_fight.weapon_value,
        _FORCED_LARGE_WEAPON_RATE=_forced_fight.large_weapon_probability,
    )

    weapon_radius = difficulty_variation(
        task_config.weapon_radius_min,
        task_config.weapon_radius_max,
        rand,
        difficulty,
        _FORCED=_forced_fight.weapon_radius,
    )

    # if smaller than this, might not necessarily be a full wal.. Is sort of related to the constant for how
    # finely meshed the worlds are
    min_wall_size = 1.0

    max_gap_dist, _warnings = world.get_critical_distance(locations, min_wall_size)

    if max_gap_dist is None:
        raise WorldTooSmall(AvalonTask.FIGHT, min_wall_size, locations.get_2d_spawn_goal_distance())

    gap_distance = normal_distrib_range(
        0.1 * max_gap_dist,
        min([max_gap_dist * 0.5, task_config.path_length_max]),
        max_gap_dist * 0.2,
        rand,
        difficulty,
        _FORCED=_forced_fight.gap_distance,
    )

    predator_size = LARGEST_ANIMAL_SIZE
    predators = []
    for i, predator_type in enumerate(predator_types):
        dist = i + 1
        predators.append(predator_type(position=np.array([-dist, predator_size / 2.0, 0.0])))

    path_width = normal_distrib_range(
        task_config.path_width_easy,
        task_config.path_width_hard,
        task_config.path_width_std_dev,
        rand,
        difficulty,
        _FORCED=_forced_fight.path_width,
    )
    ring_config = make_ring(
        rand,
        difficulty,
        world,
        locations,
        gap_distance=gap_distance,
        constraint=constraint,
        height=difficulty_variation(task_config.height_min, task_config.height_max, rand, difficulty),
        traversal_width=path_width * 2,
        # don't want you to be able to just avoid the enemy
        is_inside_climbable=False,
        is_outside_climbable=False,
        # path
        dual_solution=HeightSolution(
            paths=(
                HeightPath(
                    is_path_climbable=True,
                    is_outside_edge_unclimbable=True,
                    is_path_restricted_to_land=True,
                    is_detailed=path_width < 3.0,
                    width=path_width,
                ),
            ),
            solution_point_brink_distance=1.0,
        ),
        # "closer to spawn"
        inner_solution=attr.evolve(
            create_inner_placeholder_solution(
                count=len(weapons),
                offset=0,  # WEAPON_HEIGHT_OFFSET,
                randomization_dist=weapon_radius,
            ),
            inside_item_radius=ITEM_FLATTEN_RADIUS,
        ),
        # "further from spawn"
        outer_solution=HeightSolution(
            outside_items=tuple(predators),
            outside_item_randomization_distance=predator_randomization_radius,
            # prevents stuff from spawning this close to a predator
            outside_item_radius=predator_size,
        ),
        traversal_noise_interpolation_multiple=4.0,
        probability_of_centering_on_spawn=0.0 if is_for_carry else 0.5,
    )

    world = world.add_height_obstacle(rand, ring_config, locations.island)
    world = world.replace_weapon_placeholders(weapons, locations.island, 0.0 if is_for_carry else ITEM_FLATTEN_RADIUS)

    return world, locations, difficulty



import struct
from typing import Any
from typing import ClassVar
from typing import Dict
from typing import Generic
from typing import Protocol
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union
from typing import cast
from typing import runtime_checkable

import attr
import gym
import numpy as np
from gym import spaces
from gym.spaces import Box

from avalon.common.errors import SwitchError
from avalon.datagen.godot_base_types import Vector3


@runtime_checkable
class ActionProtocol(Protocol):
    @classmethod
    def to_gym_space(cls) -> spaces.Space:
        ...

    @classmethod
    def from_input(cls: Type["ActionType"], input_dict: Dict[str, np.ndarray]) -> "ActionType":
        ...

    def to_bytes(self) -> bytes:
        """Convert this action to bytes to be sent to godot"""

    @classmethod
    def _from_bytes_with_remainder(cls: Type["ActionType"], action_bytes: bytes) -> Tuple["ActionType", bytes]:
        """Parse this action from bytes from an action log, including unused bytes in the return"""

    @classmethod
    def from_bytes(cls: Type["ActionType"], action_bytes: bytes) -> "ActionType":
        """Parse this action from bytes from an action log"""
        return cls._from_bytes_with_remainder(action_bytes)[0]

    @classmethod
    def get_null_action(cls: Type["ActionType"]) -> "ActionType":
        """An empty action that is sent with reset messages"""


ActionType = TypeVar("ActionType", bound=ActionProtocol)
ActionType2 = TypeVar("ActionType2", bound=ActionProtocol)


class AttrsAction(ActionProtocol):
    @classmethod
    def to_gym_space(cls) -> spaces.Space:
        """Basic float space the same dimensions as this action.


        Override if any customization is necessary (i.e. discrete spaces)
        """
        field_counts = len(attr.fields(cls))
        return spaces.Box(
            low=-np.ones((field_counts,), dtype=np.float32) / 10,
            high=np.ones((field_counts,), dtype=np.float32) / 10,
            shape=(field_counts,),
            dtype=np.float32,
        )

    @classmethod
    def from_input(cls: Type[ActionType], input_dict: Dict[str, np.ndarray]) -> ActionType:
        raise NotImplementedError

    @classmethod
    def get_null_action(cls: Type[ActionType]) -> ActionType:
        action_fields = [x for x in attr.fields(cls)]
        null_action_kwargs = {}
        for field in action_fields:
            if field.type == int:
                value: Union[int, float] = 0
            elif field.type == float:
                value = 0.0
            else:
                raise SwitchError(field.type)
            null_action_kwargs[field.name] = value

        return cls(**null_action_kwargs)

    def to_bytes(self) -> bytes:
        action_bytes = b"".join(
            _to_bytes(
                cast(Type, field.type),
                getattr(self, field.name),
            )
            for field in attr.fields(self.__class__)
        )
        return _to_bytes(int, len(action_bytes)) + action_bytes

    @classmethod
    def _from_bytes_with_remainder(cls: Type[ActionType], action_bytes: bytes) -> Tuple[ActionType, bytes]:
        fields: Dict[str, Any] = {}
        _size, remaining_bytes = _from_bytes(int, action_bytes)
        for field in attr.fields(cls):
            fields[field.name], remaining_bytes = _from_bytes(cast(Type, field.type), remaining_bytes)
        return cls(**fields), remaining_bytes


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class VRAction(AttrsAction):
    """All of these that are floats have range (-1, 1). They are rescaled to physical units in the simulator."""

    head_x: float
    head_y: float
    head_z: float
    head_pitch: float
    head_yaw: float
    head_roll: float
    left_hand_x: float
    left_hand_y: float
    left_hand_z: float
    left_hand_pitch: float
    left_hand_yaw: float
    left_hand_roll: float
    right_hand_x: float
    right_hand_y: float
    right_hand_z: float
    right_hand_pitch: float
    right_hand_yaw: float
    right_hand_roll: float
    is_left_hand_grasping: float
    is_right_hand_grasping: float
    is_jumping: float

    @classmethod
    def to_gym_space(cls) -> spaces.Space:
        return spaces.Dict(
            {
                "real": Box(low=-1, high=1, shape=(18,)),
                "discrete": gym.spaces.MultiBinary(3),
            }
        )

    @classmethod
    def from_input(cls, input_dict: Dict[str, np.ndarray]) -> "VRAction":
        # clipping each triplet to sphere
        input_real = input_dict["real"]
        triplet_norm = np.linalg.norm(np.reshape(input_real, (6, 3)), axis=-1)
        scale = 1 / np.clip(triplet_norm, a_min=1, a_max=float("inf"))
        clipped_real = np.repeat(scale, 3) * input_real
        input_vec = np.concatenate([clipped_real, input_dict["discrete"]], axis=-1)
        action = cls(*tuple(x.item() for x in input_vec))
        return action


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class MouseKeyboardAction(AttrsAction):
    head_x: float
    head_z: float
    head_pitch: float
    head_yaw: float
    is_left_hand_grasping: float
    is_right_hand_grasping: float
    is_left_hand_throwing: float
    is_right_hand_throwing: float
    is_jumping: float
    is_eating: float
    is_crouching: float

    @classmethod
    def to_gym_space(cls) -> spaces.Space:
        # TODO: move this back to 2 spaces for perf gains
        return spaces.Dict(
            {
                "real": Box(low=-1, high=1, shape=(4,)),
                "discrete": gym.spaces.MultiBinary(7),
            }
        )

    @classmethod
    def from_input(cls, input_dict: Dict[str, np.ndarray]) -> "MouseKeyboardAction":
        clipped_real = np.clip(input_dict["real"], a_min=-1, a_max=1)
        input_vec = np.concatenate([clipped_real, input_dict["discrete"]], axis=-1)

        return cls(*tuple(x.item() for x in input_vec))


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class CombinedAction(ActionProtocol, Generic[ActionType, ActionType2]):

    actions: Tuple[ActionType, ActionType2]

    SPACE_DICT_PREFIX: ClassVar = "combined_action"

    @classmethod
    def get_action_types(
        cls: "Type[CombinedAction[ActionType, ActionType2]]",
    ) -> Tuple[Type[ActionType], Type[ActionType2]]:
        "Get the action types, in order, to combine"
        raise ValueError(f"Inheritors must implement get_action_types themselves")

    @classmethod
    def _get_action_types_downcast_to_satisfy_mypy(
        cls: "Type[CombinedAction[ActionType, ActionType2]]",
    ) -> Tuple[Type[ActionProtocol], Type[ActionProtocol]]:
        # mypy can't tell tuple iterations should be unions :(
        a, b = cls.get_action_types()
        assert issubclass(a, ActionProtocol)
        assert issubclass(b, ActionProtocol)
        return (a, b)

    @classmethod
    def _prefixed_space_key(cls, key: str, index: int) -> str:
        return f"{cls.SPACE_DICT_PREFIX}_{index}_{str}"

    @classmethod
    def to_gym_space(cls: "Type[CombinedAction[ActionType, ActionType2]]") -> spaces.Space:
        action_types = cls._get_action_types_downcast_to_satisfy_mypy()
        space = {}
        for i, action_type in enumerate(action_types):
            action_space = action_type.to_gym_space()
            assert isinstance(
                action_space, spaces.Dict
            ), "to_gym_space support for CombinedActions not represented as dicts is not yet supported"
            space.update({cls._prefixed_space_key(key, i): value for key, value in action_space.items()})

        return spaces.Dict(space)

    @classmethod
    def from_input(
        cls: "Type[CombinedAction[ActionType, ActionType2]]", input_dict: Dict[str, np.ndarray]
    ) -> "CombinedAction[ActionType, ActionType2]":
        action_types = cls._get_action_types_downcast_to_satisfy_mypy()
        actions = []
        for i, a_type in enumerate(action_types):
            prefix = f"{cls.SPACE_DICT_PREFIX}_{i}_"
            i_input_dict = {k[2:]: v for k, v in input_dict.items() if k.startswith(prefix)}
            actions.append(a_type.from_input(i_input_dict))
        return cls(cast(Tuple[ActionType, ActionType2], tuple(actions)))

    @classmethod
    def get_null_action(
        cls: "Type[CombinedAction[ActionType, ActionType2]]",
    ) -> "CombinedAction[ActionType, ActionType2]":
        action_types = cls._get_action_types_downcast_to_satisfy_mypy()
        null_actions = tuple(a_type.get_null_action() for a_type in action_types)
        return cls(cast(Tuple[ActionType, ActionType2], null_actions))

    def _just_field_bytes(self, action: ActionProtocol) -> bytes:
        size, field_bytes = _from_bytes(int, action.to_bytes())
        return field_bytes

    def to_bytes(self) -> bytes:
        # mypy can't tell tuple iterations should be unions
        actions = (cast(ActionProtocol, a) for a in self.actions)
        all_action_bytes = b"".join((self._just_field_bytes(a) for a in actions))
        return _to_bytes(int, len(all_action_bytes)) + all_action_bytes

    @classmethod
    def _from_bytes_with_remainder(
        cls: "Type[CombinedAction[ActionType, ActionType2]]", action_bytes: bytes
    ) -> Tuple["CombinedAction[ActionType, ActionType2]", bytes]:
        action_types = cls._get_action_types_downcast_to_satisfy_mypy()
        remaining_bytes = action_bytes
        actions = []
        for a_type in action_types:
            action, remaining_bytes = a_type._from_bytes_with_remainder(remaining_bytes)
            actions.append(action)
        final_actions = cast(Tuple[ActionType, ActionType2], tuple(actions))
        return cls(final_actions), remaining_bytes


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True, slots=True)
class DebugCameraAction(AttrsAction):
    offset: Vector3 = Vector3(0, 0, 0)
    rotation: Vector3 = Vector3(0, 0, 0)
    is_facing_tracked: bool = False
    # Setting to False will result in an empty response for the first frame,
    # but will not interfere with physics or the results of surounding actions (i.e. when replaying)
    is_frame_advanced: bool = True
    tracking_node: str = "physical_head"

    @classmethod
    def isometric(
        cls,
        tracking_node: str = "physical_head",
        distance: float = 12,
        is_facing_tracked: bool = True,
        is_frame_advanced: bool = True,
    ):
        return DebugCameraAction(
            Vector3(distance / 2, abs(distance), distance / 2),
            Vector3(np.radians(-45), np.radians(45), 0),
            is_facing_tracked,
            is_frame_advanced,
            tracking_node,
        )

    @classmethod
    def level(
        cls,
        tracking_node: str = "physical_head",
        distance: float = 12,
        is_facing_tracked: bool = True,
        is_frame_advanced: bool = True,
    ):
        return DebugCameraAction(
            Vector3(distance / 2, 0, distance / 2),
            Vector3(np.radians(-45), 0, 0),
            is_facing_tracked,
            is_frame_advanced,
            tracking_node,
        )

    def to_bytes(self) -> bytes:
        action_bytes = b""
        for vec in [self.offset, self.rotation]:
            action_bytes += _to_bytes(float, vec.x)
            action_bytes += _to_bytes(float, vec.y)
            action_bytes += _to_bytes(float, vec.z)
        action_bytes += _to_bytes(float, 1.0 if self.is_facing_tracked else 0.0)
        action_bytes += _to_bytes(float, 1.0 if self.is_frame_advanced else 0.0)
        action_bytes += self.tracking_node.encode("UTF-8")
        return _to_bytes(int, len(action_bytes)) + action_bytes

    @classmethod
    def _read_vec(cls, action_bytes: bytes) -> Tuple[Vector3, bytes]:
        v = [0.0, 1.0, 2.0]
        remaining = action_bytes
        for axis in v:
            v[axis], remaining = _from_bytes(float, action_bytes)  # type: ignore[call-overload]
        return Vector3(*v), remaining

    @classmethod
    def from_bytes(cls, action_bytes: bytes) -> "DebugCameraAction":
        _size, remaining_bytes = _from_bytes(int, action_bytes)
        offset, remaining_bytes = cls._read_vec(remaining_bytes)
        rotation, remaining_bytes = cls._read_vec(remaining_bytes)
        is_facing_tracked, remaining_bytes = _from_bytes(float, remaining_bytes)
        is_frame_advanced, remaining_bytes = _from_bytes(float, remaining_bytes)
        tracking_node = remaining_bytes.decode("UTF-8")
        return cls(offset, rotation, is_facing_tracked != 0, is_frame_advanced != 0, tracking_node)


def _to_bytes(value_type: Type, value: Any) -> bytes:
    if value_type is float:
        return struct.pack("<f", value)
    if value_type is int:
        return struct.pack("<i", value)
    raise SwitchError(f"{value_type} (value={value})")


def _from_bytes(value_type: Type, value_bytes: bytes) -> Tuple[Union[int, float], bytes]:
    byte_format = ""
    if value_type is float:
        byte_format = "<f"
    elif value_type is int:
        byte_format = "<i"
    else:
        raise SwitchError(f"{value_type} should be either float or int")

    size = struct.calcsize(byte_format)
    bytes_to_consume, bytes_remaining = value_bytes[:size], value_bytes[size:]
    return struct.unpack(byte_format, bytes_to_consume)[0], bytes_remaining



import os
import tarfile
from pathlib import Path
from typing import Optional
from typing import cast

import boto3
from botocore import UNSIGNED
from botocore.config import Config
from botocore.exceptions import ClientError

from avalon.contrib.aws_auth import load_aws_keys

REGION = "us-west-2"
TEMP_BUCKET_NAME = "untitled-ai-temp"


class SimpleS3Client:
    def __init__(self, bucket_name: str = TEMP_BUCKET_NAME, region: str = REGION) -> None:
        self.bucket_name = bucket_name
        self.region = region
        access_key: Optional[str] = None
        secret_key: Optional[str] = None
        if "AWS_ACCESS_KEY_ID" in os.environ:
            access_key, secret_key = os.environ["AWS_ACCESS_KEY_ID"], os.environ["AWS_SECRET_ACCESS_KEY"]
        else:
            access_key, secret_key = load_aws_keys()

        # Allow anonymous access, e.g. when the bucket is public
        config = None
        if access_key is None and secret_key is None:
            config = Config(signature_version=UNSIGNED)

        # Create a new session to make client creation thread-safe.
        self.client = boto3.session.Session().client(
            "s3",
            region_name=region,
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            config=config,
        )

    def save(self, key: str, data: bytes) -> None:
        self.client.put_object(
            Bucket=self.bucket_name,
            Key=key,
            Body=data,
        )

    def load(self, key: str) -> bytes:
        try:
            data = self.client.get_object(Bucket=self.bucket_name, Key=key)["Body"].read()
        except ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchKey":
                raise KeyError(key) from e
            else:
                raise
        else:
            return cast(bytes, data)  # type: ignore

    def download_to_file(self, key: str, output_path: Path) -> None:
        self.client.download_file(self.bucket_name, key, str(output_path.absolute()))

    def upload_from_file(self, input_path: Path, key: str, is_overwrite_allowed: bool = False) -> None:
        # TASK 20097119-dfe7-4abb-bdb1-3895db6742e5: remove overwrite ability from S3
        assert is_overwrite_allowed or not self.is_present(
            key
        ), f"Tried to upload to {key} but that path already exists in bucket {self.bucket_name}. Go delete that file manually (or more likely realize you did something wrong and fix that). Technically the file you tried to upload was {str(input_path)}"
        self.client.upload_file(str(input_path.absolute()), self.bucket_name, key)

    def copy_object(self, src: str, dst: str, is_overwrite_allowed: bool = False) -> None:
        # TASK 20097119-dfe7-4abb-bdb1-3895db6742e5: remove overwrite ability from S3
        assert is_overwrite_allowed or not self.is_present(
            dst
        ), f"Tried to copy s3:{src} to s3:{dst} but the destination already exists in bucket {self.bucket_name}."

        self.client.copy_object(
            CopySource={
                "Bucket": self.bucket_name,
                "Key": src,
            },
            Bucket=self.bucket_name,
            Key=dst,
        )

    def is_present(self, path: str) -> bool:
        try:
            self.client.head_object(Bucket=self.bucket_name, Key=path)
        except ClientError as e:
            if e.response["ResponseMetadata"]["HTTPStatusCode"] == 404:
                return False
            raise
        else:
            return True

    def make_bucket(self) -> None:
        self.client.create_bucket(
            Bucket=self.bucket_name,
            CreateBucketConfiguration={"LocationConstraint": self.region},  # type: ignore[typeddict-item]
        )


def download_tar_from_s3_and_unpack(tar_path: Path, client: SimpleS3Client) -> Path:
    client.download_to_file(key=tar_path.name, output_path=tar_path)
    tar = tarfile.open(tar_path)
    tar.extractall(path=tar_path.parent)
    tar.close()
    tar_path.unlink()
    return tar_path.parent



import subprocess
from typing import List
from typing import Set

from mypy.plugin import MethodSigContext
from mypy.plugin import Plugin
from mypy.types import CallableType
from mypy.types import Instance

LOCAL_CLASS_NAMES: Set[str] = set()


class TorchForwardInferencePlugin(Plugin):
    def get_method_signature_hook(self, fullname: str):
        if fullname.startswith("torch.nn.modules.") and fullname.endswith(".__call__"):
            return fix_call_method
        if fullname in LOCAL_CLASS_NAMES:
            return fix_call_method
        return None


def fix_call_method(ctx: MethodSigContext) -> CallableType:
    caller = ctx.type
    if isinstance(caller, Instance):
        if "forward" in caller.type.names:
            # go get the forward method instead
            forward_method = caller.type.names["forward"].type
            if forward_method and isinstance(forward_method, CallableType):
                if caller.type.has_base("pytorch_lightning.core.lightning.LightningModule"):
                    return forward_method
                else:
                    # and trim the "self" argument off of it
                    return forward_method.copy_modified(
                        arg_types=forward_method.arg_types[1:],
                        arg_kinds=forward_method.arg_kinds[1:],
                        arg_names=forward_method.arg_names[1:],
                    )
    return ctx.default_signature


# noinspection PyUnusedLocal
def plugin(version: str):
    global LOCAL_CLASS_NAMES

    # I'm a bad person. This is a terrible way of getting all of the classes that define a forward method...
    class_lines = subprocess.getoutput(
        'grep -E -r --include="*.py" "^(    def forward|class )" . | grep "    def forward" -B 1 | grep "class " | grep -v "mypy_torch_plugin.py"'
    ).split("\n")
    classes: List[str] = []
    for class_line in class_lines:
        try:
            file_path, class_def = class_line.split(".py:class ")
        except ValueError:
            # Sometimes that grep command gets stuff that doesn't match
            continue
        module_path = file_path[2:].replace("/", ".")
        class_name = class_def.split("(")[0]
        classes.append(f"{module_path}.{class_name}")

    LOCAL_CLASS_NAMES = {f"{x}.__call__" for x in classes}

    return TorchForwardInferencePlugin



import os
import shutil
import tarfile
import time
from collections import defaultdict
from pathlib import Path
from threading import Thread
from typing import DefaultDict
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import attrs
import numpy as np
import torch
import wandb
from matplotlib import pyplot as plt
from matplotlib.pyplot import bar
from numpy.typing import NDArray

from avalon.agent.common import wandb_lib
from avalon.agent.common.params import Params
from avalon.agent.common.storage import EpisodeStorage
from avalon.agent.common.types import Algorithm
from avalon.agent.common.types import StepData
from avalon.agent.common.worker import RolloutManager
from avalon.agent.godot.godot_gym import GodotEnvironmentParams
from avalon.agent.ppo.params import PPOParams
from avalon.agent.ppo.ppo_types import PPOStepData
from avalon.common.log_utils import logger
from avalon.contrib.s3_utils import TEMP_BUCKET_NAME
from avalon.contrib.s3_utils import SimpleS3Client
from avalon.contrib.utils import TEMP_DIR
from avalon.contrib.utils import create_temp_file_path
from avalon.datagen.world_creation.constants import int_to_avalon_task

BIG_SEPARATOR = "-" * 80
RESULT_TAG = "DATALOADER:0 TEST RESULTS"


def log_rollout_stats_packed(packed_rollouts: Dict[str, NDArray], i: int) -> None:
    """Log stats, when rollouts is a dict of packed BatchSequence tensors."""
    successes: dict[str, dict[str, list]] = defaultdict(lambda: defaultdict(list))
    keys = ["success", "difficulty"]
    if packed_rollouts["done"].sum() == 0:
        return
    for worker, timestep in np.argwhere(packed_rollouts["done"]).T:
        info = packed_rollouts["info"]
        task = int_to_avalon_task[int(info["task"][worker, timestep])].lower()
        for field in keys:
            successes[task][f"final_{field}"].append(info[field][worker, timestep])
    # Data is a dict (task) of dicts (keys) of lists
    for task, x in successes.items():
        for field, y in x.items():
            wandb_lib.log_histogram(f"train/{task}/{field}", y, i, hist_freq=10)
        wandb_lib.log_scalar(f"train/{task}/num_episodes", len(y), i)


Episode = List[StepData]
DifficultyBin = Tuple[float, float]  # [from, to)
EpisodesByDifficulty = DefaultDict[DifficultyBin, List[Episode]]
EpisodesByTaskByDifficulty = DefaultDict[str, EpisodesByDifficulty]


def get_difficulty_bin_name(difficulty_bin: DifficultyBin) -> str:
    return f"{difficulty_bin[0]:.2f}_to_{difficulty_bin[1]:.2f}"


def get_episode_difficulty_bin(episode: Episode, difficulty_bin_size: float) -> DifficultyBin:
    difficulty_bins = np.arange(difficulty_bin_size, 1 + difficulty_bin_size, difficulty_bin_size)  # interval ends
    difficulty = episode[-1].info["difficulty"]
    difficulty_bin_idx = np.digitize(difficulty, difficulty_bins, right=True)
    difficulty_bin_end = difficulty_bins[difficulty_bin_idx]
    difficulty_bin_start = difficulty_bin_end - difficulty_bin_size
    difficulty_bin = difficulty_bin_start, difficulty_bin_end
    return difficulty_bin


def log_video_by_difficulty(
    episode: Episode, difficulty_bin: DifficultyBin, prefix: str = "test", infix: str = ""
) -> None:
    if infix != "":
        infix += "/"
    difficulty_bin_name = get_difficulty_bin_name(difficulty_bin)
    video = torch.stack([step.observation["rgbd"] for step in episode])
    wandb_lib.log_video(f"{prefix}/videos/{infix}{difficulty_bin_name}", video, step=None, normalize=True, freq=1)


def log_success_by_difficulty(
    successes_by_difficulty: DefaultDict[DifficultyBin, List[int]], prefix: str = "test", suffix: Optional[str] = None
) -> None:
    if suffix is not None:
        suffix = f"/{suffix}"
    data = []
    for difficulty_bin, successes in successes_by_difficulty.items():
        difficulty_bin_name = get_difficulty_bin_name(difficulty_bin)
        success_rate = np.mean(successes)
        data.append((difficulty_bin_name, success_rate))
    data = sorted(data, key=lambda item: item[0])
    difficulty_bins, success_rates = zip(*data)
    fig = plt.figure()
    plt.ylabel("success rate")
    plt.xlabel("difficulty bin")
    plt.ylim((0, 1))
    bar(difficulty_bins, success_rates)
    wandb.log({f"{prefix}/success_by_difficulty{suffix}": wandb.Image(fig)})


def load_worlds_from_s3(data_key: str, target_path: Path, bucket_name: str = TEMP_BUCKET_NAME) -> int:
    """Download pre-generated worlds from a S3 tarball to a local directory."""
    logger.debug("started loading fixed worlds from s3 file")
    shutil.rmtree(str(target_path), ignore_errors=True)
    target_path.mkdir(parents=True)
    s3_client = SimpleS3Client(bucket_name)
    logger.info(f"Downloading data from {data_key} to {target_path}")
    with create_temp_file_path() as temp_file_path:
        s3_client.download_to_file(data_key, temp_file_path)
        with tarfile.open(temp_file_path, "r:gz") as f:
            f.extractall(target_path)
    # TODO: figure out a better way to count episodes?
    episode_count = len(os.listdir(str(target_path)))
    logger.debug("finished loading fixed worlds from s3 file")
    return episode_count


def test(params: Params, model: Algorithm, log: bool = True, log_extra: Optional[Dict[str, float]] = None):
    """Run evaluation for Godot."""
    params = attrs.evolve(params, env_params=attrs.evolve(params.env_params, mode="test"))
    # We have to pull env_params out to make mypy happy - type narrowing doesn't seem to work on instance attributes
    env_params = params.env_params
    assert isinstance(env_params, GodotEnvironmentParams)
    assert env_params.env_index == 0

    start_time = time.monotonic()

    if env_params.fixed_worlds_s3_key:
        # Load worlds from S3, if we have that enabled
        # e.g. "a5101fb5fca577a35a0749ba45ae28006823136f/test_worlds.tar.gz"
        # the worlds typically have to be put in a specific folder because they contain absolute paths...
        fixed_worlds_path = env_params.fixed_worlds_load_from_path
        if not fixed_worlds_path:
            fixed_worlds_path = Path(TEMP_DIR) / "eval_worlds"
        num_worlds = load_worlds_from_s3(env_params.fixed_worlds_s3_key, fixed_worlds_path)
        env_params = attrs.evolve(env_params, fixed_worlds_load_from_path=fixed_worlds_path)
        params = attrs.evolve(params, env_params=env_params)
    elif env_params.fixed_worlds_load_from_path:
        # Got a path but no s3 key, assume the files are already locally available
        num_worlds = len(os.listdir(str(env_params.fixed_worlds_load_from_path)))
    else:
        num_worlds = env_params.test_episodes_per_task * env_params.num_tasks

    model.reset_state()
    model.eval()

    # Set up hooks for extracting episode info we care about
    difficulty_bin_size = env_params.eval_difficulty_bin_size
    seen_worlds: set[int] = set()
    success_by_task_and_difficulty_bin: DefaultDict[str, DefaultDict[DifficultyBin, list[int]]] = defaultdict(
        lambda: defaultdict(list)
    )
    world_scores = {}
    video_logging_threads = []

    def collect_episode_stats(episode: Episode) -> None:
        world_index = episode[-1].info["world_index"]
        if world_index in seen_worlds:
            # We may run the same world twice, don't count them twice!
            logger.debug(f"got dupe world {world_index}")
            return
        else:
            seen_worlds.add(world_index)
        task = int_to_avalon_task[int(episode[-1].info["task"])].lower()
        success = episode[-1].info["success"]
        difficulty_bin = get_episode_difficulty_bin(episode, difficulty_bin_size)
        success_by_task_and_difficulty_bin[task][difficulty_bin].append(success)
        is_first_episode_in_group = len(success_by_task_and_difficulty_bin[task][difficulty_bin]) == 1
        if log and is_first_episode_in_group:
            # Note: wandb.log is not thread safe, this might cause issues. But it's fast :)
            thread = Thread(
                target=log_video_by_difficulty, args=(episode, difficulty_bin), kwargs={"infix": task}, daemon=True
            )
            thread.start()
            video_logging_threads.append(thread)
            # log_video_by_difficulty(episode, difficulty_bin, infix=task)

        # Stuff for collecting scores
        world_scores[world_index] = episode[-1].info["score"]

        # difficulty = episode[-1].info["difficulty"]
        # success = episode[-1].info["success"]
        # episode_length = len(episode)
        # logger.info(f"ep with difficulty {difficulty} had success {success} in {episode_length} steps")

    num_workers = min(params.eval_workers, num_worlds)

    if isinstance(params, PPOParams):
        step_data_type = PPOStepData
    else:
        step_data_type = StepData

    test_storage = EpisodeStorage(
        params,
        step_data_type,
        episode_callback=collect_episode_stats,
        num_workers=num_workers,
        discard_short_eps=False,
        in_memory_buffer_size=0,
    )

    # Maybe i should make a "free-running", "n_steps", and "n_episodes" worker.
    # And use the n_episodes version for eval.
    multiprocessing_context = torch.multiprocessing.get_context("spawn")
    assert params.observation_space is not None
    player = RolloutManager(
        params=params,
        num_workers=num_workers,
        is_multiprocessing=True,
        storage=test_storage,
        obs_space=params.observation_space,
        model=model,
        rollout_device=torch.device("cuda:0"),
        multiprocessing_context=multiprocessing_context,
    )
    test_storage.reset()

    try:
        logger.info(f"running {num_worlds} evaluation episodes")
        # Will potentially run some worlds multiple times
        player.run_rollout(
            num_episodes=int(np.ceil(num_worlds / num_workers)), exploration_mode=params.eval_exploration_mode
        )
        logger.debug("finished rollout, shutting down workers")
    finally:
        player.shutdown()

    [thread.join() for thread in video_logging_threads]
    end_time = time.monotonic()
    test_log: dict[str, float] = {"test_time": end_time - start_time}
    if log_extra is not None:
        test_log.update(log_extra)
    total_episodes_logged = 0
    all_successes: list[float] = []
    for task, success_by_difficulty_bin in success_by_task_and_difficulty_bin.items():
        if log:
            log_success_by_difficulty(success_by_difficulty_bin, suffix=task)
        task_successes: list[int] = sum(list(success_by_difficulty_bin.values()), [])
        all_successes.extend(task_successes)
        test_log[f"{task}_success_rate"] = float(np.mean(task_successes))
        for difficulty_bin, successes in success_by_difficulty_bin.items():
            total_episodes_logged += len(successes)
    test_log["overall_success_rate"] = float(np.mean(all_successes))
    logger.info(success_by_task_and_difficulty_bin)
    assert (
        total_episodes_logged >= num_worlds
    ), f"Expected to log at least {num_worlds}, but only logged {total_episodes_logged}"

    if log:
        wandb.log({f"test/{k}": v for k, v in test_log.items()})
        logger.info(BIG_SEPARATOR)
        logger.info(RESULT_TAG)
        logger.info(test_log)
        logger.info(BIG_SEPARATOR)

    # if env_params.fixed_worlds_load_from_path:
    #     # Special logging for the fixed evaluation worlds
    #     project = params.resume_from_project if params.resume_from_project else params.project
    #     # TODO: make this get the current wandb run_id or some other identifier if we're not loading a run
    #     run_id = params.resume_from_run
    #     filename = params.resume_from_filename
    #     fixed_world_key = env_params.fixed_worlds_s3_key if env_params.fixed_worlds_s3_key else "test"
    #     result_key = f"avalon_eval__{project}_{run_id}_{filename}__{fixed_world_key}__final"
    #     record_data = {
    #         "wandb_run": f"{project}/{run_id}/{filename}",
    #         "baseline": "PPO",
    #         "data_key": fixed_world_key,
    #         "all_results": world_scores,
    #     }
    #     logger.debug(record_data)
    #
    #     logger.info(f"Saving result to '{result_key}'")
    #     s3_client = SimpleS3Client()
    #     s3_client.save(result_key, json.dumps(record_data).encode())

    return test_log



import math
from typing import Optional
from typing import Tuple

import attr
import numpy as np

from avalon.datagen.world_creation.types import HeightMode
from avalon.datagen.world_creation.types import MapFloatNP
from avalon.datagen.world_creation.types import Point2DNP
from avalon.datagen.world_creation.worlds.obstacles.harmonics import EdgeConfig
from avalon.datagen.world_creation.worlds.obstacles.height_solution import HeightSolution


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class HeightObstacle:
    is_inside_ring: bool
    edge_config: EdgeConfig
    traversal_length: float
    # the traversal width, ie, as you are facing it, how wide (left to right) is the area that you can cross
    traversal_width: float
    # over how many multiples of traversal_width should we interpolate from "no noise" to "full noise"
    traversal_noise_interpolation_multiple: float
    is_default_climbable: bool
    detail_radius: float
    # not particularly useful... too easy for things to end up off the map
    traversal_theta_offset: float = 0.0


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class RingObstacleConfig:
    height: float
    height_mode: HeightMode
    edge: EdgeConfig
    center_point: Point2DNP
    traversal_point: Point2DNP
    inner_safety_radius: float
    outer_safety_radius: float
    inner_obstacle: Optional[HeightObstacle] = None
    outer_obstacle: Optional[HeightObstacle] = None
    inner_solution: Optional[HeightSolution] = None
    outer_solution: Optional[HeightSolution] = None
    dual_solution: Optional[HeightSolution] = None
    chasm_bottom_size: float = 0.0
    expansion_meters: float = 0.0
    terrain_blurs: Tuple[Tuple[float, float], ...] = tuple()

    def get_inner_solution_points(self) -> Tuple[Point2DNP, Point2DNP]:
        obstacle = self.inner_obstacle
        assert obstacle is not None
        delta = self.traversal_point - self.center_point
        r = np.linalg.norm(delta)
        theta = math.atan2(delta[1], delta[0])
        theta += obstacle.traversal_theta_offset
        inner_r = r - max([obstacle.traversal_length, 0.01])
        outer_r = r
        inner_point = np.array([inner_r * math.cos(theta), inner_r * math.sin(theta)])
        outer_point = np.array([outer_r * math.cos(theta), outer_r * math.sin(theta)])
        return inner_point + self.center_point, outer_point + self.center_point

    def get_outer_solution_points(self) -> Tuple[Point2DNP, Point2DNP]:
        obstacle = self.outer_obstacle
        assert obstacle is not None
        delta = self.traversal_point - self.center_point
        r = np.linalg.norm(delta)
        theta = math.atan2(delta[1], delta[0])
        theta += obstacle.traversal_theta_offset
        inner_r = r + self.chasm_bottom_size
        outer_r = r + self.chasm_bottom_size + max([obstacle.traversal_length, 0.01])
        # logger.debug(inner_r)
        # logger.debug(outer_r)
        # logger.debug(obstacle.traversal_length)
        # logger.debug(self.chasm_bottom_size)
        inner_point = np.array([inner_r * math.cos(theta), inner_r * math.sin(theta)])
        outer_point = np.array([outer_r * math.cos(theta), outer_r * math.sin(theta)])
        return inner_point + self.center_point, outer_point + self.center_point


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class RingObstacle:
    config: RingObstacleConfig
    mid_z: float
    z: MapFloatNP
    r: MapFloatNP
    theta: MapFloatNP
    outer_mid_z: float



import math
from pathlib import Path
from typing import List
from typing import Tuple

import attr
import networkx as nx
import numpy as np
from matplotlib import pyplot as plt
from scipy import stats

from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.world_creation.configs.building import BuildingConfig
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import IndoorTaskConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import AGENT_HEIGHT
from avalon.datagen.world_creation.constants import FOOD_HOVER_DIST
from avalon.datagen.world_creation.debug_plots import IS_DEBUG_VIS
from avalon.datagen.world_creation.entities.constants import FOOD_TREE_VISIBLE_HEIGHT
from avalon.datagen.world_creation.entities.constants import TREE_FOOD_OFFSET
from avalon.datagen.world_creation.entities.entity import Entity
from avalon.datagen.world_creation.indoor.builders import DefaultHallwayBuilder
from avalon.datagen.world_creation.indoor.builders import DefaultStoryLinker
from avalon.datagen.world_creation.indoor.builders import FootprintBuilder
from avalon.datagen.world_creation.indoor.builders import HouseLikeRoomBuilder
from avalon.datagen.world_creation.indoor.builders import RectangleFootprintBuilder
from avalon.datagen.world_creation.indoor.builders import TLShapeFootprintBuilder
from avalon.datagen.world_creation.indoor.builders import WindowBuilder
from avalon.datagen.world_creation.indoor.building import Building
from avalon.datagen.world_creation.indoor.building import BuildingAestheticsConfig
from avalon.datagen.world_creation.indoor.building import BuildingNavGraph
from avalon.datagen.world_creation.indoor.building import BuildingTask
from avalon.datagen.world_creation.indoor.constants import Azimuth
from avalon.datagen.world_creation.indoor.task_generator import CANONICAL_BUILDING_LOCATION
from avalon.datagen.world_creation.indoor.task_generator import BuildingTaskGenerator
from avalon.datagen.world_creation.indoor.task_generator import IndoorTaskParams
from avalon.datagen.world_creation.indoor.task_generator import create_building_for_skill_scenario
from avalon.datagen.world_creation.indoor.task_generator import get_room_centroid_in_building_space
from avalon.datagen.world_creation.indoor.task_generator import make_indoor_task_world
from avalon.datagen.world_creation.indoor.task_generator import rectangle_dimensions_within_radius
from avalon.datagen.world_creation.tasks.eat import add_food_tree_for_simple_task
from avalon.datagen.world_creation.types import WorldType
from avalon.datagen.world_creation.worlds.creation import create_world_for_skill_scenario
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.export import export_world


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class ExploreIndoorTaskConfig(IndoorTaskConfig):
    # Largest possible site radius for a building to be placed on at difficulty=1
    max_site_radius: float = 25.0
    # Standard deviation for the distribution deciding the site radius: higher means more variability at same difficulty
    site_radius_std_dev: float = 2.5
    # Min/max number of stories used for the building at difficulty=0 and 1, respectively
    min_story_count: int = 1
    max_story_count: int = 3


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class ExploreTaskConfig(TaskConfig):
    # likelihood that the indoor variant of the task will be used
    indoor_probability: float = 0.2
    # max distance to make any of the compositional tasks. In meters.
    outdoor_max_distance: float = 600.0
    indoor_config: ExploreIndoorTaskConfig = ExploreIndoorTaskConfig()


def generate_explore_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: ExploreTaskConfig = ExploreTaskConfig(),
) -> None:
    is_indoor = rand.uniform() < task_config.indoor_probability
    if is_indoor:
        building, entities, spawn_location, target_location = create_building_for_skill_scenario(
            rand,
            difficulty,
            ExploreTaskGenerator(task_config.indoor_config),
            position=CANONICAL_BUILDING_LOCATION,
            is_indoor_only=True,
        )
        world = make_indoor_task_world(
            building, entities, difficulty, spawn_location, target_location, rand, export_config
        )
    else:
        desired_goal_dist = scale_with_difficulty(difficulty, 0.5, task_config.outdoor_max_distance / 2.0)
        world, locations = create_world_for_skill_scenario(
            rand,
            difficulty,
            TREE_FOOD_OFFSET,
            stats.norm(desired_goal_dist, desired_goal_dist / 5),
            export_config,
            is_visibility_required=False,
            visibility_height=FOOD_TREE_VISIBLE_HEIGHT,
            max_size_in_meters=task_config.outdoor_max_distance,
            world_type=WorldType.CONTINENT,
        )
        world = add_food_tree_for_simple_task(world, locations)
        world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal, is_visibility_required=False)

    export_world(output_path, rand, world)


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class ExploreTaskGenerator(BuildingTaskGenerator):
    config: ExploreIndoorTaskConfig = ExploreIndoorTaskConfig()

    def get_site_radius(self, rand: np.random.Generator, difficulty: float) -> float:
        return normal_distrib_range(
            self.config.min_site_radius, self.config.max_site_radius, self.config.site_radius_std_dev, rand, difficulty
        )

    def get_building_config(
        self,
        rand: np.random.Generator,
        difficulty: float,
        radius: float,
        allowed_auxiliary_tasks: Tuple[BuildingTask, ...] = tuple(),
        aesthetics: BuildingAestheticsConfig = BuildingAestheticsConfig(),
    ) -> BuildingConfig:
        story_count = self.config.min_story_count + round(
            normal_distrib_range(0, self.config.max_story_count - self.config.min_story_count, 0.1, rand, difficulty)
        )
        width, length = rectangle_dimensions_within_radius(radius)
        footprint_builder: FootprintBuilder
        if width <= 10 or length <= 10:
            footprint_builder = RectangleFootprintBuilder()
            min_room_size = 2
        else:
            footprint_builder = TLShapeFootprintBuilder()
            min_room_size = 3

        return BuildingConfig(
            width=width,
            length=length,
            story_count=story_count,
            footprint_builder=footprint_builder,
            room_builder=HouseLikeRoomBuilder(min_room_size=min_room_size),
            hallway_builder=DefaultHallwayBuilder(proportion_additional_edges=1),
            story_linker=DefaultStoryLinker(
                raise_on_failure=False, allow_ladders=BuildingTask.CLIMB in allowed_auxiliary_tasks
            ),
            window_builder=WindowBuilder(),
            aesthetics=aesthetics,
        )

    def get_principal_obstacle_params(self, rand: np.random.Generator, difficulty: float, building: Building) -> Tuple:
        spawn_story = building.stories[0]
        if len(spawn_story.rooms) == 1:
            raise ImpossibleWorldError("Building too small for explore task")

        exterior_wall_azimuths_by_room_id = spawn_story.get_exterior_wall_azimuths_by_room_id()
        periphery_room_ids = [
            room_id
            for room_id, exterior_wall_azimuths in exterior_wall_azimuths_by_room_id.items()
            if len(exterior_wall_azimuths) > 0
        ]
        initial_room_id = rand.choice(periphery_room_ids)

        nav_graph = BuildingNavGraph(building)
        spawn_room = spawn_story.rooms[initial_room_id]
        spawn_node = nav_graph.get_room_node(spawn_story, spawn_room)
        distance_by_location_id = nx.single_source_dijkstra_path_length(nav_graph, spawn_node, weight="distance")
        max_distance = max(distance_by_location_id.values())
        desired_distance = difficulty * max_distance

        # The target location distribution uses a standard deviation derived from the actual distribution derivation,
        # since otherwise you can end up with all-zero weights for sparse/bimodal distributions
        # (e.g. [10, 100] at difficulty 0.5 = 50, for which weights are [0,0] if we use a hard-coded std of say 0.5
        distribution_std = np.std(list(distance_by_location_id.values()))
        target_location_distribution = stats.norm(desired_distance, math.sqrt(distribution_std))
        location_weights = np.array(
            [
                target_location_distribution.pdf(distance) if distance > 0 else 0
                for distance in distance_by_location_id.values()
            ]
        )
        location_weights /= location_weights.sum()
        target_node = rand.choice(list(distance_by_location_id.keys()), p=location_weights)
        return nav_graph, spawn_node, target_node

    def add_principal_obstacles(
        self, rand: np.random.Generator, building: Building, obstacle_params: Tuple
    ) -> IndoorTaskParams:
        nav_graph, spawn_node, target_node = obstacle_params
        stories_and_rooms_by_node_id = nav_graph.get_stories_and_rooms_by_node_id(building.stories)

        spawn_story, spawn_room = stories_and_rooms_by_node_id[spawn_node]
        spawn_location = get_room_centroid_in_building_space(
            building, spawn_story, spawn_room, at_height=AGENT_HEIGHT / 2
        )
        target_location = np.array(nav_graph.nodes[target_node]["position"])
        target_location[1] += FOOD_HOVER_DIST

        if IS_DEBUG_VIS:
            ax = nav_graph.plot()
            ax.text(
                spawn_location[2], spawn_location[0], spawn_location[1] - (AGENT_HEIGHT // 2), "spawn", color="blue"
            )
            ax.text(target_location[2], target_location[0], target_location[1], "target", color="green")
            plt.show()

        extra_items: List[Entity] = []
        entrance_sites = ((spawn_story.num, spawn_room.id, tuple(Azimuth)),)
        return building, extra_items, spawn_location, target_location, entrance_sites



from __future__ import annotations

from typing import Dict
from typing import Optional

import gym
import numpy as np
import torch
import torch.nn as nn
from einops import rearrange
from torch import Tensor
from torch.distributions import Independent
from torch.distributions import Normal
from torch.distributions import OneHotCategorical
from torch.distributions.transformed_distribution import TransformedDistribution
from torch.nn import functional as F

from avalon.agent.common import wandb_lib
from avalon.agent.common.params import ClippedNormalMode
from avalon.agent.common.params import Params
from avalon.agent.common.types import ActionBatch
from avalon.agent.common.wrappers import OneHotMultiDiscrete
from avalon.agent.dreamer.tools import SampleDist
from avalon.agent.dreamer.tools import TanhBijector
from avalon.agent.dreamer.truncated_normal import ClippedTruncatedNormal


class NormalWithMode(Normal):
    def mode(self) -> Tensor:
        mean = self.mean
        assert isinstance(mean, Tensor)
        return mean

    def __getitem__(self, item):
        # slicing
        assert isinstance(item, (int, tuple))
        return NormalWithMode(self.loc[item], self.scale[item])


class IndependentWithMode(Independent):
    def mode(self) -> Tensor:
        mode = self.base_dist.mode()
        assert isinstance(mode, Tensor)
        return mode

    def __getitem__(self, item):
        # slicing
        assert isinstance(item, (int, tuple))
        if isinstance(item, tuple):
            assert len(item) == 1  # this could be lifted
        assert self.reinterpreted_batch_ndims == 1
        # if this was an Independent(1), then slicing just removes this wrapper?
        return self.base_dist[item]


class NormalHead(nn.Module):
    """A module that builds a Diagonal Gaussian distribution from means.

    If model_provides_std=False, standard deviations are learned parameters in this module.
    Otherwise they are taken as inputs.
    """

    def __init__(
        self,
        action_space: gym.spaces.Box,
        init_std: float,
        min_std: float,
        clipped_normal_mode: ClippedNormalMode,
        model_provides_std: bool = True,
    ) -> None:
        super().__init__()
        assert isinstance(action_space, gym.spaces.Box)
        assert len(action_space.shape) == 1
        assert np.all(action_space.low == -1.0)
        assert np.all(action_space.high == 1.0)
        self.num_outputs = action_space.shape[0]
        self.model_provides_std = model_provides_std
        self.min_std = min_std
        self.clipped_normal_mode = clipped_normal_mode
        if model_provides_std:
            # We take 2 values per output element, one for mean and one for (raw) std
            self.num_inputs = self.num_outputs * 2
            self.init_std = init_std
        else:
            # # We'll use a constant learned std
            # self.num_inputs = self.num_outputs
            # # initial variance is e^0 = 1
            # self.log_std = nn.Parameter(torch.zeros(self.num_outputs))
            raise NotImplementedError

    def forward(self, x: Tensor) -> torch.distributions.Distribution:
        # x should have shape (..., action_dim)
        assert x.shape[-1] == self.num_inputs
        if self.model_provides_std:
            mean, raw_std = torch.chunk(x, 2, -1)
        else:
            raise NotImplementedError
            # mean = x
            # std = self.log_std.exp() + self.min_std

        if self.clipped_normal_mode == ClippedNormalMode.NO_CLIPPING:
            # note that standard practice would be to use std = log_std.exp(), here we use a centered softplus instead
            raw_init_std = np.log(np.exp(self.init_std) - 1)  # such that init_std = softplus(raw_init_std)
            std = F.softplus(raw_std + raw_init_std) + self.min_std
            dist: torch.distributions.Distribution = NormalWithMode(mean, std)
            return IndependentWithMode(dist, reinterpreted_batch_ndims=1)
        elif self.clipped_normal_mode == ClippedNormalMode.SAMPLE_DIST:
            # This is exactly matching the dreamerv1 (need to double check) and v2-"tanh-normal" mode
            # this mean_scale and tanh thing is a sort of soft clipping of the input.
            mean_scale = 5
            mean_scaled = mean_scale * torch.tanh(mean / mean_scale)
            # note that standard practice would be to use std = log_std.exp(), here we use a centered softplus instead
            raw_init_std = np.log(np.exp(self.init_std) - 1)  # such that init_std = softplus(raw_init_std)
            std = F.softplus(raw_std + raw_init_std) + self.min_std
            dist = Normal(mean_scaled, std)
            transformed_dist = TransformedDistribution(dist, TanhBijector())
            independent_dist = Independent(transformed_dist, 1)
            sample_dist = SampleDist(independent_dist)
            return sample_dist
        elif self.clipped_normal_mode == ClippedNormalMode.TRUNCATED_NORMAL:
            # This is exactly the default mode for dreamerv2.
            # These are the dreamer default params
            # assert self.init_std == 0
            # assert self.min_std == 0.1
            # To be precise, we should inverse-sigmoid init_std, but this is how dreamerv2 did it.
            std = 2 * torch.sigmoid((raw_std + self.init_std) / 2) + self.min_std
            dist = ClippedTruncatedNormal(torch.tanh(mean), std, -1, 1)
            dist = IndependentWithMode(dist, 1)
            return dist
        assert False


class StraightThroughOneHotCategorical(OneHotCategorical):
    def rsample(self, sample_shape: torch.Size = torch.Size()):
        assert sample_shape == torch.Size()
        # Straight through biased gradient estimator.
        sample = self.sample(sample_shape).to(torch.float32)
        probs = self.probs
        assert sample.shape == probs.shape
        sample += probs - probs.detach()
        return sample.float()

    def mode(self) -> Tensor:
        return F.one_hot(self.probs.argmax(dim=-1), self.event_shape[-1]).float()  # type: ignore

    def __getitem__(self, item):
        # slicing
        assert isinstance(item, (int, tuple))
        return StraightThroughOneHotCategorical(self.logits[item])


class MultiCategoricalHead(nn.Module):
    """Represents multiple categorical dists. All must have the same number of categories."""

    def __init__(self, num_actions: int, num_categories: int, center_and_clamp_logits: bool = False) -> None:
        super().__init__()
        self.num_categories = num_categories
        self.num_actions = num_actions
        self.num_inputs = num_categories * num_actions
        self.num_outputs = num_categories * num_actions
        self.center_and_clamp_logits = center_and_clamp_logits

    def forward(self, x: Tensor) -> torch.distributions.Distribution:
        assert x.shape[-1] == self.num_actions * self.num_categories
        x = rearrange(x, "... (a c) -> ... a c", a=self.num_actions, c=self.num_categories)
        if self.center_and_clamp_logits:
            # This was found to consistently hurt performance across PPO Procgen envs,
            # but was used (and presumably helped at some point) in Avalon
            x = x - x.mean(dim=-1, keepdim=True)
            x = torch.clamp(x, -4, 4)
        dist = StraightThroughOneHotCategorical(logits=x)
        # TODO: i'm not sure this is right if there's multiple actions in this MultiCategorical dist.
        independent_dist = IndependentWithMode(dist, 1)
        return independent_dist


class DictActionHead(torch.nn.Module):
    """This model handles generating a policy distribution from latents.

    Latents should be passed in with shape (..., self.num_inputs), and a DictActionDist will be returned.
    """

    def __init__(self, action_space: gym.spaces.Dict, params: Params) -> None:
        super().__init__()
        self.action_space = action_space

        # Build action heads
        self.num_inputs = 0
        self.num_outputs = 0
        action_heads = {}
        for k, space in self.action_space.spaces.items():
            if isinstance(space, gym.spaces.Box):
                # The input is (..., num_actions)
                # The output of sample() should be (..., num_actions)
                head: torch.nn.Module = NormalHead(
                    space,
                    init_std=params.policy_normal_init_std,
                    min_std=params.policy_normal_min_std,
                    clipped_normal_mode=params.clipped_normal_mode,
                    model_provides_std=params.normal_std_from_model,
                )
            elif isinstance(space, OneHotMultiDiscrete):
                # A MultiDiscrete is multiple independent Discrete spaces.
                # We coerce all discrete space types into this with a wrapper
                # This won't work if the discretes have different num_categories.
                assert len(set(space.nvec)) == 1
                head = MultiCategoricalHead(
                    num_actions=len(space.nvec),
                    num_categories=space.max_categories,
                    center_and_clamp_logits=params.center_and_clamp_discrete_logits,
                )
            else:
                assert False
            action_heads[k] = head
            self.num_inputs += action_heads[k].num_inputs  # type: ignore
            self.num_outputs += action_heads[k].num_outputs  # type: ignore
        self.action_heads = torch.nn.ModuleDict(action_heads)

    def forward(self, action_logits: Tensor) -> "DictActionDist":
        dists = {}
        i = 0
        for k, head in self.action_heads.items():
            logits = action_logits[..., i : i + head.num_inputs]
            dists[k] = head(logits)
            i += head.num_inputs
        return DictActionDist(dists)


class DictActionDist(torch.distributions.Distribution):
    """This is an instance of a torch Distribution that holds key-value pairs of other Distributions.

    It's used for e.g. sampling from an entire Dict action space in one operation,
    which will return a dictionary of samples.

    Operations like entropy() will reduce over all dists to return a single value (per batch element)."""

    def __init__(self, dists: dict[str, torch.distributions.Distribution]) -> None:
        super().__init__(validate_args=False)
        self.dists = dists

    def sample(self) -> ActionBatch:  # type: ignore
        # Detach is just for good measure in case someone messes up the sample() impl :)
        return {k: v.sample().detach() for k, v in self.dists.items()}

    def rsample(self) -> ActionBatch:  # type: ignore
        return {k: v.rsample() for k, v in self.dists.items()}

    def log_prob(self, actions: dict[str, Tensor]) -> Tensor:
        """Compute the log prob of the given action under the given dist (batchwise).

        Log prob is a scalar (per batch element.)"""
        # actions is a dict of tensors of shape (batch_size, num_outputs)
        log_probs = []
        for k, dist in self.dists.items():
            # batch_size = actions[k].shape[0]
            # assert actions[k].shape == (batch_size, self.action_heads[k].num_outputs)
            log_prob = dist.log_prob(actions[k])
            # The log_prob is over the entire action space, so it reduces to a single scalar per batch element
            # assert log_prob.shape == (batch_size,)
            log_probs.append(log_prob)

        # the output should have shape (batch_size, )
        return torch.stack(log_probs, dim=1).sum(dim=1)

    def entropy(self) -> Tensor:
        # The entropy is over the entire action space, so it reduces to a single scalar per batch element
        entropies = [v.entropy() for v in self.dists.values()]
        return torch.stack(entropies, dim=1).sum(dim=1)

    def mean(self) -> Dict[str, Tensor]:
        return {k: v.mean() for k, v in self.dists.items()}

    def mode(self) -> Dict[str, Tensor]:
        return {k: v.mode() for k, v in self.dists.items()}

    def __getitem__(self, item: str):
        if isinstance(item, str):
            # user is treating this as a dict
            return self.dists[item]
        elif isinstance(item, (int, tuple)):
            # user is slicing in the batch dim
            return DictActionDist({k: v[item] for k, v in self.dists.items()})
        else:
            raise ValueError

    def __iter__(self):
        # in case anyone tries to iterate on this (cough cough wandb.watch)
        return self.dists.values()


def visualize_action_dists(
    action_space: gym.spaces.Dict,
    action_dist_dict: DictActionDist,
    prefix: str = "action_dists",
    freq: Optional[int] = None,
):
    """Log the action dists to wandb. Any batch shape is fine.

    The action dists often give more info than the actual actions, since the actions are sampled from these dists.
    """
    for k, space in action_space.spaces.items():
        action_dist = action_dist_dict[k]
        if isinstance(space, OneHotMultiDiscrete):
            # We coerce all discrete spaces into this type now
            # action_probs should have shape (batch_dims, num_actions, num_categories)
            # The dist here is an IndependentWithMode(StraightThroughOneHotCategorical())
            probs = action_dist.base_dist.probs
            probs = probs.reshape(-1, probs.shape[-2], probs.shape[-1])
            for act_i in range(len(space.nvec)):
                for cat_i in range(int(space.nvec[act_i])):
                    wandb_lib.log_histogram(
                        f"{prefix}/{k}_{act_i}_{cat_i}", probs[:, act_i, cat_i], mean_freq=freq, hist_freq=freq
                    )
                    if space.nvec[act_i] == 2:
                        # a binary space only needs one of the categories logged
                        break
        elif isinstance(space, gym.spaces.Box):
            assert len(space.shape) == 1
            if isinstance(action_dist, SampleDist):
                raise NotImplementedError
            elif isinstance(action_dist, IndependentWithMode):
                action_dist = action_dist.base_dist
                assert isinstance(action_dist, (ClippedTruncatedNormal, NormalWithMode))
                means = action_dist.mean
                for dim in range(means.shape[-1]):
                    wandb_lib.log_histogram(
                        f"{prefix}/{k}_{dim}_mean", means[..., dim], mean_freq=freq, hist_freq=freq
                    )
                stds = action_dist.stddev
                for dim in range(means.shape[-1]):
                    wandb_lib.log_histogram(f"{prefix}/{k}_{dim}_std", stds[..., dim], mean_freq=freq, hist_freq=freq)


def visualize_actions(
    action_space: gym.spaces.Dict, action_dict: dict[str, Tensor], prefix: str = "actions", freq: Optional[int] = None
) -> None:
    """Log a batch of actions to wandb (any batch shape is fine)."""
    for k, space in action_space.spaces.items():
        action = action_dict[k]
        if isinstance(space, OneHotMultiDiscrete):
            # expect one-hot of shape (batch_dims, num_actions, num_categories)
            # flatten the batch dim
            action = action.reshape(-1, action.shape[-2], action.shape[-1])
            for act_i in range(len(space.nvec)):
                for cat_i in range(int(space.nvec[act_i])):
                    # Histograms don't make sense, we'll just log the % of actions where this category is selected
                    wandb_lib.log_scalar(
                        f"{prefix}/{k}_{act_i}_{cat_i}", action[:, act_i, cat_i].float().mean(), freq=freq
                    )
                    if space.nvec[act_i] == 2:
                        # a binary space only needs one of the categories logged
                        break
        elif isinstance(space, gym.spaces.Box):
            assert len(space.shape) == 1
            # expect shape (batch_dims, num_actions)
            for act_i in range(space.shape[0]):
                wandb_lib.log_histogram(
                    f"{prefix}/{k}_{act_i}", action[:, act_i].float(), mean_freq=freq, hist_freq=freq
                )



from typing import Final

LARGEST_ANIMAL_SIZE: Final = 3.125
STICK_LENGTH: Final = 4.0
FOOD_TREE_BASE_HEIGHT = 2.0
CANONICAL_FOOD_HEIGHT_ON_TREE = 2.0
FOOD_TREE_VISIBLE_HEIGHT = 11.0
TREE_FOOD_OFFSET = 2.0



from typing import cast

import attr
import numpy as np

from avalon.common.utils import to_immutable_array
from avalon.datagen.world_creation.types import MapBoolNP
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.utils import to_2d_point


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class WorldLocations:
    island: MapBoolNP = attr.ib(converter=to_immutable_array)
    spawn: Point3DNP = attr.ib(converter=to_immutable_array)
    goal: Point3DNP = attr.ib(converter=to_immutable_array)

    def get_2d_spawn_goal_distance(self) -> float:
        return cast(float, np.linalg.norm(to_2d_point(self.spawn) - to_2d_point(self.goal)))
